indexing

	description:
		"Reads an XML document and attempts to retrieve all DTD specifications from it."


class

	DTD_PARSER


inherit

	XM_EXPAT_PARSER
		rename
			make as make_expat
		redefine
			on_xml_declaration,
			on_element_declaration,
			on_attribute_declaration,
			on_doctype
		end

	EXCEPTIONS

	KL_SHARED_PLATFORM

	GF_SCIENTIFIC_FORMAT_ACCESSOR


creation

	make


feature

	make (a_schema: SCHEMA_SPECIFICATION) is
		require
			schema_not_void: a_schema /= Void
		do
			make_expat
			schema := a_schema
		end


feature -- handlers

	on_xml_declaration (xml_version, encoding: UC_STRING; is_standalone: BOOLEAN) is
		do
			register_all_callbacks
			enable_parameter_entity_parsing
		end

	on_doctype (name: STRING; an_id: XM_DTD_EXTERNAL_ID; has_internal_subset: BOOLEAN) is
		do
			schema.set_root_element (name)
		end

	on_element_declaration (a_name: STRING; a_model: XM_DTD_ELEMENT_CONTENT) is
		local
			element_children: DEFINED_ELEMENT_CHILDREN
			element_empty: DEFINED_ELEMENT_EMPTY
			element_mixed: DEFINED_ELEMENT_MIXED
			child_def_mixed: CHILD_DEFINITION_MIXED
			empty_child_list: DS_LINKED_LIST [CHILD_DEFINITION]
		do
			debug
				print ("<!ELEMENT ")
				print (a_name)
				print (">%N")
			end
			if a_model.is_content_mixed then
				if a_model.items.count > 0 then
					create element_mixed.make (a_name, make_child_definition (a_model))
				else
					if a_model.items.count = 0 then
						create empty_child_list.make
						create child_def_mixed.make (empty_child_list)
						create element_mixed.make (a_name, child_def_mixed)
					else
						create element_mixed.make (a_name, make_child_definition (a_model))
					end
				end
				schema.elements.put_last (element_mixed)
			elseif a_model.is_choice or else a_model.is_sequence then
				create element_children.make (a_name, make_child_definition (a_model))
				schema.elements.put_last (element_children)
			elseif a_model.is_content_empty then
				create element_empty.make (a_name)
				schema.elements.put_last (element_empty)
			else
				print ("Unhandeld model encountered.%N")
			end

--  			type := exml_XML_cp_type (model)
--  			if type = XML_CTYPE_CHOICE or else type = XML_CTYPE_SEQ then
--  				create element_children.make (name, make_child_definition (model))
--  				schema.elements.put_last (element_children)
--  			elseif type = XML_CTYPE_EMPTY then
--  				create element_empty.make (name)
--  				schema.elements.put_last (element_empty)
--  			elseif type = XML_CTYPE_MIXED then
--  				if exml_XML_cp_numchildren (model) > 0 then
--  					create element_mixed.make (name, make_child_definition (model))
--  				else
--  					if exml_XML_cp_numchildren (model) = 0 then
--  						create empty_child_list.make
--  						create child_def_mixed.make (empty_child_list)
--  						create element_mixed.make (name, child_def_mixed)
--  					else
--  						create element_mixed.make (name, make_child_definition (model))
--  					end
--  				end
--  				schema.elements.put_last (element_mixed)
--  			else
--  				print ("Other types (!d) not yet supported.%N")
--  			end
		end

	on_attribute_declaration (an_element_name, a_name: STRING; a_model: XM_DTD_ATTRIBUTE_CONTENT) is
		do
		end


feature {NONE} -- child definition parser

	make_child_definition (model: XM_DTD_ELEMENT_CONTENT): CHILD_DEFINITION is
		require
			valid_model: model /= Void
			not_leaf: not model.is_name
			there_are_children: model.items.count > 0
		do
			Result := make_child_definition_occurrence (model)
		ensure
			valid_result: Result /= Void
		end

	make_child_definition_type (model: XM_DTD_ELEMENT_CONTENT): CHILD_DEFINITION is
		require
			model_not_void: model /= Void
		local
			childs: DS_LINKED_LIST [CHILD_DEFINITION]
		do
			if model.is_name then
				create {CHILD_DEFINITION_NAME} Result.make (model.name)
			elseif model.is_choice then
				childs := make_child_list (model)
				create {CHILD_DEFINITION_CHOICE} Result.make (childs)
			elseif model.is_sequence then
				childs := make_child_list (model)
				create {CHILD_DEFINITION_SEQ} Result.make (childs)
			elseif model.is_content_mixed then
				childs := make_child_list (model)
				create {CHILD_DEFINITION_MIXED} Result.make (childs)
			else
				raise ("Unhandled model type should not occur%N.")
			end
		ensure
			valid_result: Result /= Void
		end

	make_child_definition_occurrence (model: XM_DTD_ELEMENT_CONTENT): CHILD_DEFINITION is
			-- make a single child definition, wrap it in a quantifier
			-- (?, +, *) if necessary
		require
			model_not_void: model /= Void
		local
			child_type: CHILD_DEFINITION
		do
			child_type := make_child_definition_type (model)

			if model.is_one then
				Result := child_type
			else
				create {CHILD_DEFINITION_OCCURRENCE} Result.make (child_type, model.minimum_occurrences, model.maximum_occurrences)
			end
		end

	make_child_list (model: XM_DTD_ELEMENT_CONTENT): DS_LINKED_LIST [CHILD_DEFINITION] is
			-- Handle every child of `model'.
		require
			valid_model: model /= Void
			not_leaf: not model.is_name
			there_are_children: model.items.count > 0
		local
			child_model: XM_DTD_ELEMENT_CONTENT
			child: CHILD_DEFINITION
		do
			create Result.make
			from
				model.items.start
			until
				model.items.after
			loop
				child_model := model.items.item_for_iteration
				child := make_child_definition_occurrence (child_model)
				Result.put_last (child)
				model.items.forth
			end
		ensure
			result_not_void: Result /= Void
		end


feature -- what I've parsed so far

	schema: SCHEMA_SPECIFICATION


invariant

	schema_not_void: schema /= Void

end
