indexing
   
   description: "Should behave as XML_EVENT_PARSER??"

   known_bugs:
   "1. Does not support name spaces."


-- deferred
class 

   ABSTRACT_XML_PARSER
   
   
inherit

   XML_EVENT_PARSER
      redefine
         on_xml_declaration,
         on_content,
         on_start_tag,
         on_end_tag,
         on_start_doctype,
         parse_from_file_name
      end

   XML_EVENT_PARSER_I
      rename
         on_content as int_content,
         on_default as int_default,
         on_default_expanded as int_default_expanded,
         on_start_tag as int_start_tag,
         on_end_tag as int_end_tag,
         on_start_doctype as int_start_doctype,
         on_processing_instruction as int_processing_instruction,
         on_comment as int_comment,
         on_end_doctype as int_end_doctype,
         on_not_standalone as int_not_standalone,
         on_start_cdata_section as int_start_cdata_section,
         on_end_cdata_section as int_end_cdata_section,
         on_entity_declaration as int_entity_declaration,
         on_element_declaration as int_element_declaration,
         on_start_namespace_declaration as int_start_namespace_declaration,
         on_end_namespace_declaration as int_end_namespace_declaration,
         on_notation_declaration as int_notation_declaration,
         on_attribute_declaration as int_attribute_declaration,
         on_xml_declaration as int_xml_declaration
      end

   G_SCIENTIFIC_FORMAT
      rename
         make as format_make,
         print as format_print
      end

   KL_SHARED_EXCEPTIONS

   
creation
   
   make
   

feature -- creation

   make is
      do
         format_make
         create elements.make (512) -- he, we're not written for trivial DTDs!
      end
   

feature -- Expat

   my_parser: EXPAT_EVENT_PARSER

   create_event_parser_imp: EXPAT_EVENT_PARSER is
         -- Our default implementation is Expat, override to use others
      do
         create Result
      end

   parse_from_file_name (a_file_name: UC_STRING) is
      do
         create state.make
         my_parser := create_event_parser_imp
         make_from_imp (my_parser)
         precursor (a_file_name)
      end


feature -- building schema definition
   
   -- there is no name space support here, perhaps put every element
   -- in its own name space to start supporting it?
   
   elements: DS_HASH_TABLE [DEFINED_ELEMENT, UC_STRING]

   add_element (element: DEFINED_ELEMENT) is
      require
         valid_element: element /= Void
      do
         elements.put (element, element.name)
      end
   
   locate_element (name: UC_STRING): DEFINED_ELEMENT is
      require
         valid_name: name /= Void and not name.empty
      do
         elements.search (name)
         if elements.found then
            Result := elements.found_item
         end
      end


feature {NONE} -- incoming tag/data handling

   frozen on_xml_declaration (xml_version, encoding: UC_STRING; standalone: INTEGER) is
      do
         my_parser.register_doctype_handler
         -- my_parser.enable_parameter_entity_parsing
      end

   frozen on_start_tag (name, 
                        ns_prefix: UC_STRING; 
                        attributes: DS_BILINEAR [DS_PAIR [DS_PAIR [UC_STRING, UCSTRING], UCSTRING]]) is
      require
         have_state: state /= Void
         have_elements: elements /= Void
      local
         next_element: DEFINED_ELEMENT
         element_state: ELEMENT_STATE
      do
         debug            
            print ("<")
            print (name.out)
            print (">%N")
         end
         
         next_element := locate_element (name)
         if next_element = Void then
            -- is not a DEFINED_ELEMENT, halt
            element_not_defined (name)
         end         
         
         if state.root = Void then
            -- check against root type set with DOCTYPE instruction
            if state.doctype_root /= Void then
               if state.doctype_root /= next_element then
                  first_element_not_doctype_root (name)
               end
            end
            state.set_root (next_element)
            create element_state.make (state.root)
            state.parents.put (element_state)
         else
            if is_valid_child (next_element) then
               
               check
                  transition_ok: state.current_element = next_element
               end
               
               precursor (name, ns_prefix, attributes)
            else
               invalid_child (state.current_element, next_element)
            end               
         end
      end
         
   frozen on_end_tag (name, ns_prefix: UC_STRING) is
      local
         element: DEFINED_ELEMENT
      do
         debug            
            print ("</")
            print (name.out)
            print (">%N")
         end

         -- this is usually catched by the parser, but in case it
         -- isn't...
         element := locate_element (name)
         if element = Void then
            -- is not a DEFINED_ELEMENT, halt
            element_not_defined (name)
         end
         if state.current_element /= element then
            unexpected_end_tag (state.current_element, element)
         end
         
         if is_valid_end (element) then
            -- do_end_tag (element)           
            precursor (name, ns_prefix)
         else
            required_child_expected (element)
         end
      end
   
   frozen on_content (chr_data: UC_STRING) is
      do
         -- also XML Schema check here to validate content against 
         -- defined base types
         if state.current_element.is_data_allowed then
            precursor (chr_data)
         else
            element_does_not_allow_data (state.current_element)
         end
      end

   frozen on_start_doctype (name, system_id, public_id: UC_STRING; has_internal_subset: BOOLEAN) is
      local
         root: DEFINED_ELEMENT
      do
         root := locate_element (name)
         if root = Void then
            element_not_defined (name)
         end
         state.set_doctype_root (root)
         int_start_doctype (name, system_id, public_id, has_internal_subset)
      end


feature -- XML didn't validate, what to do?
   
   element_not_defined (tag_name: UC_STRING) is
      do
         validation_error (sprintf ("Tag <!s> is undefined.", <<tag_name.out>>))
      end

   first_element_not_doctype_root (tag_name: UC_STRING) is
      do
         validation_error (sprintf ("First tag <!s> is not equal to the expected root.", <<tag_name.out>>))
      end

   element_does_not_allow_data (element: DEFINED_ELEMENT) is
      do
         validation_error (sprintf ("Element <!s> does not allow character data.", <<element.name.out>>))
      end

   invalid_child (parent, child: DEFINED_ELEMENT) is
      do
         if expected_element = Void then
            validation_error (sprintf ("Parent <!s> does not allow child <!s> at the current position.", <<parent.name.out, child.name.out>>))
         else
            validation_error (sprintf ("Parent <!s> expects <!s> at the current position, but got <!s>.", <<parent.name.out, expected_element.name.out, child.name.out>>))
         end
      end

   unexpected_end_tag (parent, element: DEFINED_ELEMENT) is
      do
         validation_error (sprintf ("Element </!s> cannot end tag <!s>.", <<parent.name.out, element.name.out>>))
      end
   
   required_child_expected (element: DEFINED_ELEMENT) is
      do
         if expected_element = Void then
            validation_error (sprintf ("Tag </!s> comes too soon. Another tag was expected.", <<element.name.out>>))
         else
            validation_error (sprintf ("Tag </!s> comes too soon. Tag <!s> expected.", <<element.name.out, expected_element.name.out >>))
         end
      end

   validation_error (error_msg: STRING) is
      do
         debug
            print ("Validation error: ")
            print (error_msg)
            print ("%N")
         end
         Exceptions.raise (error_msg)
      end


feature -- error reporting state
   
   expected_element: DEFINED_ELEMENT is
      do
         Result := state.current_element_state.expected_element
      end


feature -- state
   
   state: PARSE_STATE



feature -- state transition   
   
   is_valid_child (next_element: DEFINED_ELEMENT): BOOLEAN is
         -- returns True if next_element is a valid child of 
         -- `current_element', regarding all previos childs that we've
         -- seen
      require
         valid_current: state.current_element /= Void
         there_is_a_parent: not state.parents.is_empty
      local
         element_state: ELEMENT_STATE
      do
         Result := state.current_element.is_valid_child (state.current_element_state, next_element)
         if Result then
            -- child says yes, this tag becomes the current one
            create element_state.make (next_element)
            state.parents.put (element_state)
         end
      ensure
         parent_stack_maintained: Result implies state.current_element = next_element
      end

   is_valid_end (element: DEFINED_ELEMENT): BOOLEAN is
      require
         is_current_element: state.current_element = element
      do
         Result := state.current_element.is_valid_end (state.current_element_state)
         if Result then
            state.parents.remove
         end
      ensure
         parent_stack_maintained: Result implies state.parents.is_empty or else state.current_element /= element
      end
   

invariant
   
   valid_elements: elements /= Void

end   
