indexing
   
   description: "Describes ?, * or +"
   

class
   
   CHILD_DEFINITION_OCCURRENCE
   
   
inherit
   
   CHILD_DEFINITION
   

creation
   
   make
   
   
feature -- creation
   
   make (a_child: CHILD_DEFINITION; a_min, a_max: INTEGER) is
      require
         valid_child: a_child /= Void
         valid_a_min: a_min >= 0
         valid_a_max: a_max >= 1 and a_max >= a_min
      do
         child := a_child
         min_occurrences := a_min
         max_occurrences := a_max
      end
   

feature -- state
   
   min_occurrences: INTEGER
   
   max_occurrences: INTEGER
   
   child: CHILD_DEFINITION
   
   
feature {CHILD_DEFINITION} -- start tag transition
   
   is_allowed_set (state: ELEMENT_STATE; 
                       next_element: DEFINED_ELEMENT): BOOLEAN is
      local
         child_state: CHILD_STATE
      do
         debug
            print ("CHILD_DEFINITION_OCCURRENCE.is_allowed_set%N")
         end
         
         -- push new state level on stack, this object becomes active
         -- done proactively, so any state state pushed by my children
         -- is top
         create child_state.make (Current)
         state.children.put (child_state)
         
         Result := child.is_allowed_set (state, next_element)
         if Result then
            child_state.advance_occurrences
         else
            state.children.remove
            -- child may think it's authorative, but for optional 
            -- occurrences, we have to overrule that decision
            if min_occurrences = 0 then
               state.set_authoritative_false (False)
            end
         end
      end
   
   is_allowed_stay (state: ELEMENT_STATE; 
                    next_element: DEFINED_ELEMENT): BOOLEAN is
      do
         debug
            print ("CHILD_DEFINITION_OCCURRENCE.is_allowed_stay%N")
         end
         
         state.current_child_state.advance_occurrences
         Result := state.current_child_state.occurrences <= max_occurrences
         if Result then
            -- attempt to set child, not allowed when we have an
            -- element that does not occur in the thing we make occur.
            Result := child.is_allowed_set (state, next_element)
            
            -- child may think it's authorative, but when we've seen
            -- our minimal required occurrences, we have to overrule
            -- that decision
            if not Result then
               if state.current_child_state.occurrences >= min_occurrences then
                  state.set_authoritative_false (False)
               end
            end

         end

         if not Result then
            state.children.remove
         end
      end


feature {CHILD_DEFINITION} -- end tag transition
   
   any_required_child (state: ELEMENT_STATE): BOOLEAN is
         -- return True if, from the position record in `state', there is
         -- a required child definition
      do
         debug
            print ("CHILD_DEFINITION_OCCURRENCE.any_required_child.%N")
         end

         -- if we have been set, we can just compare against the
         -- minimum number of occurrences
         if state.child_already_set (Current) then
            Result := min_occurrences > state.current_child_state.occurrences
         else
            Result := min_occurrences > 0
         end
      end


end
