indexing

	description: "Describes element type children."


class

	DEFINED_ELEMENT_CHILDREN


inherit

	DEFINED_ELEMENT
		rename
			make as inherited_make,
			make_ns as inherited_make_ns
		redefine
			init_element_state
		end


creation

	make,
	make_empty_ns,
	make_ns


feature {NONE} -- Initialization

	make (a_name: STRING; a_child_definition: CHILD_DEFINITION) is
		require
			name_not_empty: a_name /= Void and then not a_name.is_empty
			child_definition_not_void: a_child_definition /= Void
		do
			inherited_make (a_name)
			create child_definition.make (a_child_definition)
		end

	make_empty_ns (a_namespace, a_name: STRING) is
		require
			name_not_empty: a_name /= Void and then not a_name.is_empty
		do
			inherited_make_ns (a_namespace, a_name)
			is_building := True
		ensure
			building: is_building
		end

	make_ns (a_namespace, a_name: STRING; a_child_definition: CHILD_DEFINITION) is
		require
			name_not_empty: a_name /= Void and then not a_name.is_empty
			child_definition_not_void: a_child_definition /= Void
		do
			inherited_make_ns (a_namespace, a_name)
			create child_definition.make (a_child_definition)
		end


feature -- State

	child_definition: CHILD_DEFINITION_ROOT

	set_child_definition (a_child_definition: CHILD_DEFINITION) is
			-- Replace children of this element. It's main use is to
			-- allow self-references to be handled easily when building a
			-- schema.
		require
			root_child_definition_void: child_definition = Void
			child_definition_not_void: a_child_definition /= Void
		do
			create child_definition.make (a_child_definition)
			is_building := False
		end


feature -- Check current tag transition

	is_valid_child (
			state: ELEMENT_STATE;
			next_element: DEFINED_ELEMENT): BOOLEAN is
		do
			Result := child_definition.is_allowed_child (state, next_element)
		end

	is_valid_end (state: ELEMENT_STATE): BOOLEAN is
		do
			Result := child_definition.is_allowed_end (state)
		end

	is_data_allowed: BOOLEAN is
		do
			--- data is not allowed
			Result := False
		end

	is_valid_data (data: STRING): BOOLEAN is
		do
			--- data is never valid
			Result := False
		end


feature {ELEMENT_STATE}

	init_element_state (state: ELEMENT_STATE) is
			-- chance to initialize optional state fields so
			-- `is_valid_child' can start with the first child
		do
			-- my child_definitions like to have this
			state.create_children
		ensure
			child_stack_available: state.children.is_empty
		end


feature {NONE} -- Implementation

	is_building: BOOLEAN
			-- Temporary state to be used when handling self-references
			-- during schema building.

invariant

	at_least_one_child: child_definition /= Void or else is_building

end
